/*
 * display.c:
 * Display images gathered by driftnet.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

#ifndef NO_DISPLAY_WINDOW

static const char rcsid[] = "$Id: display.c,v 1.15 2002/06/13 20:06:42 chris Exp $";

#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include <time.h>

#include <sys/stat.h>

#include "driftnet.h"

/* The border, in pixels, around images displayed in the window. */
#define BORDER  6

extern int verbose; /* in driftnet.c */

static GtkWidget *window, *darea;
static GdkWindow *drawable;

static int width, height;

int *lower_bounds = NULL;

struct imgrect
  {
    GdkRectangle  rect;
    char         *filename;
    GdkPixbuf    *pixbuf;
  };

static int nimgrects;
static struct imgrect *imgrects;

gint
delete_event (GtkWidget *widget, GdkEvent *event, gpointer data)
{
    if (verbose)
        fprintf(stderr, PROGNAME ": display child shutting down\n");
    return FALSE;   /* do destroy window */
}

/* update_window:
 * Copy the backing image onto the window. */
void
update_window ()
{
  gtk_widget_queue_draw (darea);
}

/* add_image_rectangle:
 * Add a rectangle representing the location of an image to the list, so that
 * we can do hit-tests against it. */
struct imgrect *
add_image_rectangle (const char *filename,
                     const int x, const int y)
{
  GError *error = NULL;
  struct imgrect *ir;

  for (ir = imgrects; ir < imgrects + nimgrects; ++ir) {
      if (!ir->filename)
          break;
  }
  if (ir == imgrects + nimgrects) {
      imgrects = realloc(imgrects, 2 * nimgrects * sizeof *imgrects);
      memset(imgrects + nimgrects, 0, nimgrects * sizeof *imgrects);
      ir = imgrects + nimgrects;
      nimgrects *= 2;
  }
  ir->filename = strdup (filename);
  if (ir->pixbuf)
    {
      gdk_pixbuf_unref (ir->pixbuf);
    }

  ir->pixbuf = gdk_pixbuf_new_from_file (ir->filename, &error);

  if (error)
    {
      g_printerr (PROGNAME ": %s\n", error->message);
      g_clear_error (&error);

      if (ir->filename)
        free (ir->filename);
      if (ir->pixbuf)
        gdk_pixbuf_unref (ir->pixbuf);
      memset (ir, 0, sizeof *ir);
      return NULL;
    }
  
  if (!ir->pixbuf)
    g_printerr ("No Pixbuf - huh?\n");

  ir->rect.width  = gdk_pixbuf_get_width (ir->pixbuf);
  ir->rect.height = gdk_pixbuf_get_height (ir->pixbuf);

  if (ir->rect.width > 300 || ir->rect.height > 300)
    {
      GdkPixbuf *scaledbuf;

      gdouble factor = 300. / ir->rect.width;

      if (rint (factor * ir->rect.height) > 250)
        factor = 300. / ir->rect.height;

      scaledbuf = gdk_pixbuf_scale_simple (ir->pixbuf,
                                           rint (factor * ir->rect.width),
                                           rint (factor * ir->rect.height),
                                           GDK_INTERP_BILINEAR);

      gdk_pixbuf_unref (ir->pixbuf);
      ir->pixbuf = scaledbuf;

      if (!ir->pixbuf)
        {
          g_printerr ("scaling failed??\n");
          if (ir->filename)
            free (ir->filename);
          if (ir->pixbuf)
            gdk_pixbuf_unref (ir->pixbuf);
          memset (ir, 0, sizeof *ir);
          return NULL;
        }

      ir->rect.width = gdk_pixbuf_get_width (ir->pixbuf);
      ir->rect.height = gdk_pixbuf_get_height (ir->pixbuf);
    }

      
  ir->rect.x = x;
  ir->rect.y = y;

  return ir;
}


void
free_image_rectangle (struct imgrect *ir)
{
  if (ir->filename)
    {
      unlink (ir->filename);
      free (ir->filename);
    }

  if (ir->pixbuf)
    gdk_pixbuf_unref (ir->pixbuf);

  memset (ir, 0, sizeof *ir);
}


void
check_image_rectangles ()
{
  struct imgrect *ir;
  
  for (ir = imgrects; ir < imgrects + nimgrects; ++ir)
    {
      if (ir->pixbuf)
        {
          /* outside the window, no longer in use */

          if ((ir->rect.y + ir->rect.height) < 0 || ir->rect.x > width)
            free_image_rectangle (ir);
        }
    }

}


/* scroll_backing_image:
 * Scroll the image up a bit, to make room for a new image. */
void
scroll_backing_image (const int dy)
{
  struct imgrect *ir;
  
  for (ir = imgrects; ir < imgrects + nimgrects; ++ir)
    if (ir->pixbuf)
      ir->rect.y -= dy;

  check_image_rectangles ();
}


/* place_image_rectangle:
 * try to find the topmost location to place a new image of a certain
 * width.
 */
void
place_image_rectangle (int r_width, int r_height, int *x, int *y)
{
  gint max = 0;
  gint i, j, h;

  /* find the topmost place with width r_width */
  for (i = 0; i + r_width < width; i++)
    {
      h = lower_bounds[i];
      for (j = 1; j < r_width; j++)
        if (lower_bounds[i+j] < h)
          h = lower_bounds[i+j];
      
      if (i == 0 || h > max)
        {
          max = h;
          *x = i;
        }
    }

  /* ensure that the image isn't placed too close to the top of the image */
  if (max > height / 3 * 2)
    max = height / 3 * 2;

  /* maybe we need to scroll? */

  if (max < r_height)
    {
      scroll_backing_image (r_height - max);
      for (i = 0; i < width; i++)
        lower_bounds[i] += r_height - max;

      max = r_height;
    }

  *y = height - max;

  /* adjust the lower boundary of the images */
  for (i = 0; i < r_width; i++)
    lower_bounds[*x + i] = max - r_height;
}


/* find_image_rectangle:
 * Find the image, if any, which contains a given point. Used for saving images
 * when they are clicked on. */
struct imgrect *
find_image_rectangle (const int x, const int y)
{
    struct imgrect *ir;

    for (ir = imgrects; ir < imgrects + nimgrects; ++ir)
      if (ir->pixbuf &&
          x >= ir->rect.x && x < ir->rect.x + ir->rect.width &&
          y >= ir->rect.y && y < ir->rect.y + ir->rect.height)
        return ir;

    return NULL;
}


/* expose_event:
 * React to an expose event, perhaps changing the backing image size. */
void
expose_event (GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
  struct imgrect *ir;
  GdkRectangle affected;

  if (darea)
      drawable = darea->window;

  gdk_draw_rectangle (drawable, darea->style->black_gc, TRUE,
                      event->area.x, event->area.y,
                      event->area.width, event->area.height);

  for (ir = imgrects; ir < imgrects + nimgrects; ++ir)
    {
      if (ir->pixbuf &&
          gdk_rectangle_intersect (&event->area, &ir->rect, &affected))
        {
          gdk_draw_pixbuf (drawable, darea->style->black_gc, ir->pixbuf,
                           affected.x - ir->rect.x,
                           affected.y - ir->rect.y,
                           affected.x, affected.y,
                           affected.width, affected.height,
                           GDK_RGB_DITHER_MAX,
                           0, 0);
        }
    }
}


/* configure_event:
 * React to a configure event
 */
void
configure_event (GtkWidget *widget, GdkEventConfigure *event, gpointer data)
{
  gint dy, i;
  gint *new_lbounds = NULL;
  gint average = 0;

  dy = height - event->height;

  new_lbounds = g_new0 (int, event->width);

  if (lower_bounds)
    {
      for (i=0; i < width && i < event->width; i++)
        {
          new_lbounds[i] = lower_bounds[i];
          average += new_lbounds[i];
        }

      average = average / width;

      g_free (lower_bounds);
    }

  if (event->width > width)
    {
      for (i = width; i < event->width; i++)
        new_lbounds[i] = average;
    }

  lower_bounds = new_lbounds;

  width = event->width;
  height = event->height;

  scroll_backing_image (dy);

  update_window();
}

/* char *savedimgpfx:
 * The filename prefix with which we save images. */
char *savedimgpfx = "driftnet-";

/* save_image:
 * Save an image which the user has selected. */
void
save_image(struct imgrect *ir)
{
    static char *name;
    static int num;
    int fd1, fd2;
    char buf[8192];
    ssize_t l;
    struct stat st;

    if (!name)
        name = calloc(strlen(savedimgpfx) + 16, 1);

    do
        sprintf(name, "%s%d%s", savedimgpfx, num++, strrchr(ir->filename, '.'));
    while (stat(name, &st) == 0);
    fprintf(stderr, PROGNAME": saving `%s' as `%s'\n", ir->filename, name);

    fd1 = open(ir->filename, O_RDONLY);
    if (fd1 == -1) {
        fprintf(stderr, PROGNAME": %s: %s\n", ir->filename, strerror(errno));
        return;
    }
    
    fd2 = open(name, O_WRONLY | O_CREAT | O_TRUNC, 0666);
    if (fd2 == -1) {
        fprintf(stderr, PROGNAME": %s: %s\n", name, strerror(errno));
        close(fd1);
        return;
    }

    /* XXX interrupts */
    while ((l = read(fd1, buf, sizeof buf)) > 0) {
        if (write(fd2, buf, l) == -1) {
            fprintf(stderr, PROGNAME": %s: %s\n", name, strerror(errno));
            close(fd1);
            close(fd2);
            return;
        }
    }

    if (l == -1)
        fprintf(stderr, PROGNAME": %s: %s\n", ir->filename, strerror(errno));

    close(fd1);
    close(fd2);
}

static struct
{
    int x, y;
} click;

void
key_press_event (GtkWidget *widget, GdkEventKey *event)
{
  static gboolean fullscreen = FALSE;

  switch (event->keyval)
  {
    case GDK_F11:
      if (fullscreen)
        gtk_window_unfullscreen (GTK_WINDOW (window));
      else
        gtk_window_fullscreen (GTK_WINDOW (window));
      fullscreen = !fullscreen;
      break;
      
    case GDK_Escape:
    case GDK_Q:
    case GDK_q:
      gtk_main_quit ();
      break;
  }
}

void
button_press_event (GtkWidget *widget, GdkEventButton *event)
{
  click.x = (int)event->x;
  click.y = (int)event->y;
}

void
button_release_event 
(GtkWidget *widget, GdkEventButton *event)
{
  struct imgrect *ir;
  ir = find_image_rectangle(click.x, click.y);
  if (ir && ir == find_image_rectangle((int)event->x, (int)event->y)) {
      /* We draw a little frame around the image while we're saving it, to
       * give some visual feedback. */
      struct timespec jiffy = { 0, 100000000 };
      gdk_draw_rectangle(drawable, darea->style->white_gc, 0, ir->rect.x - 2, ir->rect.y - 2, ir->rect.width + 3, ir->rect.height + 3);
      gdk_flush();    /* force X to actually draw the damn thing. */
      save_image(ir);
      nanosleep(&jiffy, NULL);
      gdk_draw_rectangle(drawable, darea->style->black_gc, 0, ir->rect.x - 2, ir->rect.y - 2, ir->rect.width + 3, ir->rect.height + 3);
  }
}

void
destroy (GtkWidget *widget, gpointer data)
{
  gtk_main_quit();
}

extern int dpychld_fd;  /* in driftnet.c */

/* xread:
 * Like read(2) but read the whole supplied length. */
static ssize_t xread(int fd, void *buf, size_t len) {
    char *p;
    for (p = (char*)buf; p < (char*)buf + len; ) {
        ssize_t l;
        l = read(fd, p, (char*)buf + len - p);
        if (l == -1 && errno != EINTR)
            return -1;
        else if (l == 0)
            return 0;
        else
            p += l;
    }
    return len;
}

/* pipe_event:
 * React to events on the connecting pipe by loading images from the temporary
 * directory and displaying them on the window. */
extern char *tmpdir;    /* in driftnet.c */

gboolean
pipe_event (GIOChannel chan, GIOCondition cond, gpointer data)
{
    static char *path;
    char name[TMPNAMELEN];
    ssize_t rr;
    int nimgs = 0;

    if (!path)
        path = malloc(strlen(tmpdir) + 34);

    /* We are sent messages continaing the length of the filename, then the
     * length of the file, then the filename. */
    while (nimgs < 4 && (rr = xread(dpychld_fd, name, sizeof name)) == sizeof name) {
        int saveimg = 0;
        struct stat st;

        ++nimgs;
        
        sprintf(path, "%s/%s", tmpdir, name);

        if (stat(path, &st) == -1)
            continue;
           
        if (verbose)
            fprintf(stderr, PROGNAME": received image %s of size %d\n", name, (int)st.st_size);
        /* Check to see whether this looks like an image we're interested in. */
        /* Small images are probably bollocks. */
        if (st.st_size > 256)
          {
            struct imgrect *ir;

            ir = add_image_rectangle (path, 0, height);

            if (ir && ir->pixbuf)
              {
                if (ir->rect.width > 8 && ir->rect.height > 8)
                  {
                    /* slot in the new image at some plausible place. */
                    int w, h;
                    if (ir->rect.width > width - 2 * BORDER)
		      w = width - 2 * BORDER;
                    else
		      w = ir->rect.width;
                    if (ir->rect.height > height - 2 * BORDER)
		      h = height - 2 * BORDER;
                    else
		      h = ir->rect.height;

                    place_image_rectangle (w+2*BORDER, h+2*BORDER, &ir->rect.x, &ir->rect.y);

                    ir->rect.width = w;
                    ir->rect.height = h;

                    ir->rect.x += BORDER;
                    ir->rect.y += BORDER;

                    saveimg = 1;

                    update_window ();

                  }
                else
                  {
                    if (verbose)
                      fprintf(stderr, PROGNAME": %s: image dimensions (%d x %d) too small to bother with\n", name, ir->rect.width, ir->rect.height);
                    free_image_rectangle (ir);
                  }
              }
          }
        else
          {
            if (verbose)
              fprintf(stderr, PROGNAME": image data too small (%d bytes) to bother with\n", (int)st.st_size);
          }

        if (!saveimg)
            unlink(name);
      }
    if (rr == -1 && errno != EINTR && errno != EAGAIN) {
        perror(PROGNAME": read");
        gtk_main_quit();
    } else if (rr == 0) {
        /* pipe closed, exit. */
        gtk_main_quit();
    }
    return TRUE;
}

int
dodisplay (int argc, char *argv[])
{
    GIOChannel *chan;
    struct imgrect *ir;

    /* have our main loop poll the pipe file descriptor */
    chan = g_io_channel_unix_new(dpychld_fd);
    g_io_add_watch(chan, G_IO_IN | G_IO_ERR | G_IO_HUP, (GIOFunc)pipe_event, NULL);
    fcntl(dpychld_fd, F_SETFL, O_NONBLOCK);

    /* set up list of image rectangles. */
    imgrects = calloc(nimgrects = 16, sizeof *imgrects);
       
    /* do some init thing */
    gtk_init(&argc, &argv);
    gdk_rgb_init();

    gtk_widget_set_default_colormap(gdk_rgb_get_cmap());
    gtk_widget_set_default_visual(gdk_rgb_get_visual());

    /* Make our own window. */
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_default_size (GTK_WINDOW (window), 600, 440);

    darea = gtk_drawing_area_new();
    gtk_container_add (GTK_CONTAINER(window), darea);
    gtk_widget_set_events (darea,
                           GDK_EXPOSURE_MASK | GDK_KEY_PRESS_MASK |
                           GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
    gtk_widget_modify_bg (darea, GTK_STATE_NORMAL, &darea->style->black);

    gtk_signal_connect(GTK_OBJECT (window), "delete_event",
                       GTK_SIGNAL_FUNC (delete_event), NULL);
    gtk_signal_connect(GTK_OBJECT (window), "destroy",
                       GTK_SIGNAL_FUNC (destroy), NULL);

    gtk_signal_connect(GTK_OBJECT (darea), "expose-event",
                       GTK_SIGNAL_FUNC (expose_event), NULL);
    gtk_signal_connect(GTK_OBJECT (darea), "configure_event",
                       GTK_SIGNAL_FUNC (configure_event), NULL);
    
    /* mouse button press/release for saving images */
    gtk_signal_connect(GTK_OBJECT (darea), "button_press_event",
                       GTK_SIGNAL_FUNC(button_press_event), NULL);
    gtk_signal_connect(GTK_OBJECT (darea), "button_release_event",
                       GTK_SIGNAL_FUNC(button_release_event), NULL);
    gtk_signal_connect(GTK_OBJECT (window), "key_press_event",
                       GTK_SIGNAL_FUNC(key_press_event), NULL);

    gtk_widget_show_all(window);

    gdk_drawable_get_size (darea->window, &width, &height);

    gtk_main();

    /* Get rid of all remaining images. */
    for (ir = imgrects; ir < imgrects + nimgrects; ++ir)
      free_image_rectangle (ir);

    g_free (lower_bounds);
    return 0;
}

#endif /* !NO_DISPLAY_WINDOW */
